<?php

/**
 * @file
 * Controller class for Lionbridge translator plugin.
 */

/**
 * Lionbridge translation plugin controller.
 */
class TMGMTLionbridgeTranslatorPluginController extends TMGMTDefaultTranslatorPluginController {

  /**
   * Implements TMGMTTranslatorPluginControllerInterface::requestTranslation().
   */
  public function requestTranslation(TMGMTJob $job) {
    switch ($job->state) {
      case TMGMT_JOB_STATE_UNPROCESSED:
        try {
          $this->submitJob($job);
          if (!$job->isRejected()) {
            $job->submitted(t('Job has been submitted.'));
          }
        }
        catch (TMGMTException $e) {
          watchdog_exception('lionbridge_translation_provider', $e);
          $job->rejected('Job has been rejected with following error: @error',
            ['@error' => $e->getMessage()], 'error');
        }
        break;

      case TMGMT_JOB_STATE_ACTIVE:
        try {
          $this->fetchJob($job);
        }
        catch (TMGMTException $e) {
          watchdog_exception('lionbridge_translation_provider', $e);
          $job->addMessage('Translation could not be completed: @error',
            ['@error' => $e->getMessage()], 'error');
        }
        break;
    }

  }

  /**
   * Implements TMGMTTranslatorPluginControllerInterface::isAvailable().
   */
  public function isAvailable(TMGMTTranslator $translator) {
    if ($translator->getSetting('access_key_id') && $translator->getSetting('access_key')) {
      return TRUE;
    }

    return FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function canTranslate(TMGMTTranslator $translator, TMGMTJob $job) {
    $jobstate = $job->getState();
    if($jobstate == TMGMT_JOB_STATE_UNPROCESSED || $jobstate == TMGMT_JOB_STATE_ABORTED || $jobstate == TMGMT_JOB_STATE_REJECTED){
      return FALSE;
    }
    return TRUE;
  }

  /**
   * TMGMTDefaultTranslatorPluginController::getSupportedRemoteLanguages().
   */
  public function getSupportedRemoteLanguages(TMGMTTranslator $translator) {
    if ($translator->isAvailable() == FALSE) {
      return parent::getSupportedRemoteLanguages($translator);
    }

    if (!empty($this->supportedRemoteLanguages)) {
      return $this->supportedRemoteLanguages;
    }

    try {
      $api_client = new TMGMTLionbridgeConnector($translator);
      $languages = $api_client->listLocales();
      if (!isset($languages['Error']) && !isset($languages['head']) && !isset($languages['body'])) {
        foreach ($languages['Locale'] as $language) {
          $this->supportedRemoteLanguages[$language['Code']] = $language['Code'];
        }
      }
    }
    catch (TMGMTException $e) {
      watchdog_exception('lionbridge_translation_provider', $e);
      drupal_set_message($e->getMessage(), 'error');
    }

    return $this->supportedRemoteLanguages;
  }

  /**
   * Submits a job to Lionbridge translation service.
   *
   * This does the work of actually submitting the translation job to the
   * Lionbridge translation service. There are 3 major parts to submitting a
   * translations job. Each part has been broken out into an individual
   * function to make the code more readable.
   *
   * 1. Create a JSON encoded file that contains all the translatable content
   *    and send it to Lionbridge. If this is successful, an array containing
   *    the details of the uploaded file is returned.
   *
   * 2. Create a project using the settings from the job settings page and the
   *    AssetID of the array returned from the successful file upload. If there
   *    were more than one file uploaded, multiple file asset ID's could be
   *    used in a project, however, in this model there is only 1 asset ID for
   *    a given project.
   *
   * 3. Generate a quote for the translation job using the Project ID from the
   *    array returned from a successful project creation request.
   *
   * @param \TMGMTJob $job
   *   The job object.
   */
  public function submitJob(TMGMTJob $job) {
    $translator = $job->getTranslator();
    $api_client = new TMGMTLionbridgeConnector($translator);
    $data_flat = array_filter(tmgmt_flatten_data($job->getData()), '_tmgmt_filter_data');
    $items = $job->getItems();
    $file_assets = $product_assets = array();
    $source_language = $translator->mapToRemoteLanguage($job->source_language);
    $target_language = $translator->mapToRemoteLanguage($job->target_language);
    $target_languages = array($target_language);

    // This is the name of the JSON file that will be sent to Lionbridge.
    $file_name = _lionbridge_translation_provider_generate_file_name($job->label, $source_language, $target_language);

    $project_complete_url = url('lionbridge_translation_provider_project_complete_callback', array('absolute' => TRUE));
    if ($translator->getSetting('notification_url')) {
      $project_complete_url = $translator->getSetting('notification_url');
    }

    /*
     * Create an array of job item metadata. This is necessary because there
     * is nothing in the tmgmt_remote table yet to lookup these job items and
     * we need a way to refer to them later.
     */
    $translation_keys = $translatable_content = array();

    switch ($translator->getSetting('job_type')) {
      case 'tm_update':
        $language_data = [
          'source_language' => $source_language,
          'target_language' => $target_language,
        ];
        $update_data = [];

        // Set the data_flat so easier to use.
        $data_items = [];
        foreach ($data_flat as $key => $value) {
          $field_parts = explode('][', $key);
          $key_tjiid = (int) array_shift($field_parts);

          $data_items[$key_tjiid][] = [
            'key' => $key,
            'field' => $field_parts,
            'label' => $value['#label'],
            'text' => $value['#text'],
          ];
        }

        // Process the job items for the submitted project.
        foreach ($items as $tjiid => $item) {
          // Load entity based on entity type.
          $entity = entity_load($item->item_type, array((int) $item->item_id));
          if (!$entity) {
            // If entity is not loaded, it means that the Job Item contained
            // a non-entity. E.g. menu, menu_link, taxonomy_vocabulary.
            // TM Feature supports entities only - throws an exception.
            // E.g. Node, User, Taxonomy Term. etc.
            throw new TMGMTException(t('TM Feature does not support "@type" type elements. You may request support on Drupal.org Lionbridge project page.', array('@type' => $item->item_type)));
          }

          $entity_path = entity_uri($item->item_type, $entity[$item->item_id]);

          if (isset($entity[$item->item_id]->tnid) && (int) $entity[$item->item_id]->tnid !== 0) {
            // If content was translated using Core's Content Translation module
            // then we assume it's a node, since only node translations will be
            // supported by TMGMT in this case.
            $translations = translation_node_get_translations($entity[$item->item_id]->tnid);

            // Process the translated node.
            $tnode = node_load((int) $translations[$job->target_language]->nid);

            if (!$tnode) {
              continue;
            }

            // Process for the translated fields related to current job item.
            $tfields = $this->getUpdatedDataCoreTranslation($entity[$item->item_id], $tnode, $data_items, $tjiid);
          }
          elseif (module_exists('entity_translation')) {
            // If Entity Translation module is enabled and content wasn't
            // translated using Content Translation module, then we'll be
            // looking for field translations.
            $languages = [
              'source' => $job->source_language,
              'target' => $job->target_language,
            ];

            $tfields = $this->getUpdatedDataEntityTranslation($entity[$item->item_id], $item->item_type, $data_items, $tjiid, $languages);
          }

          if (!count($tfields)) {
            continue;
          }

          $update_data[] = [
            'entity_path' => $entity_path['path'],
            'fields' => $tfields,
          ];
        }

        $batch_size = 500;
        $file_count = round(count($update_data) / $batch_size, 0, PHP_ROUND_HALF_DOWN);
        for ($i = 0; $i <= $file_count; $i++) {
          $offset = $batch_size * $i;
          $update_batch = array_slice($update_data, $offset, $batch_size);

          // Set the file content using translated data array.
          // Max 500 job items per file.
          $file_content = theme('lionbridge_translation_provider_tmupdate', [
            'language_data' => $language_data,
            'item_data' => $update_batch,
          ]);

          // Add the file for the XML.
          $translation = [
            'source_language' => $source_language,
            'fileName' => urlencode($file_name . '-' . $i . '-' . '.xml'),
            'content_type' => 'text/xml',
            'file_content' => $file_content,
          ];

          $file_asset = $this->addFile($translation, $api_client);
          if (isset($file_asset['Errors']['Error'])) {
            throw new TMGMTException($file_asset['Errors']['Error']['DetailedMessage']);
          }

          $file_assets[] = $file_asset['AssetID'];
        }
        break;

      default:
        foreach ($data_flat as $key => $value) {
          list($tjiid, $data_item_key) = explode('][', $key, 2);

          // Preserve the original data.
          $translation_keys[] = [
            'tjiid' => $tjiid,
            'data_item_key' => $data_item_key,
          ];

          // This is the actual text sent to Lionbridge for translation.
          $translatable_content[$tjiid][$data_item_key] = $value['#text'];
        }
        $file_content = json_encode($translatable_content, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE);
        $file_content = str_replace(['“', '”'], '\"', $file_content);
        $translation = [
          'source_language' => $source_language,
          'fileName' => urlencode($file_name . '.json'),
          'content_type' => 'application/json',
          'file_content' => $file_content,
        ];

        $file_asset = $this->addFile($translation, $api_client);
        if (isset($file_asset['Errors']['Error'])) {
          throw new TMGMTException($file_asset['Errors']['Error']['DetailedMessage']);
        }

        $file_assets[] = $file_asset['AssetID'];
        break;
    }

    // Create a new project with the file asset, send it to Lionbridge. This
    // will create a new project and return the project XML in the form of an
    // array.
    $project_data = [
      'project_title' => $job->settings['project_title'] . ' to ' . $target_language,
      'service_id' => $translator->getSetting('service'),
      'source_language' => $source_language,
      'target_languages' => $target_languages,
      'product_assets' => $product_assets,
      'file_assets' => $file_assets,
    ];

    $project = $this->addProject($project_data, $api_client);

    if (isset($project['Errors']['Error']) && !empty($project['Errors']['Error'])) {
      throw new TMGMTException($project['Errors']['Error']['DetailedMessage']);
    }

    // Get a quote for the project.
    $quote = $this->generateQuote($project['ProjectID'], $project_complete_url, $api_client);

    if (isset($quote['Errors']) && !empty($quote['Errors'])) {
      throw new TMGMTException($quote['Errors']['Error']['DetailedMessage']);
    }

    // Set the job reference to the Quote ID.
    $job->reference = $quote['QuoteID'];
  }

  /**
   * Get structured source and translated data using Core Translation method.
   *
   * @param object $node
   *   Source entity.
   * @param object $tnode
   *   Translated entity.
   * @param array $data_items
   *   Job item data items.
   * @param int $tjiid
   *   Translation Job Item ID.
   *
   * @return array
   *   Translated fields data.
   */
  public function getUpdatedDataCoreTranslation($node, $tnode, $data_items, $tjiid) {
    $tfields = array();

    foreach ($data_items[$tjiid] as $row) {
      $field_name = $row['field'][0];

      switch ($field_name) {
        case 'node_title':
          $field_name = 'title';
          $translated_content = $tnode->title;
          $source_content = $node->title;
          break;

        default:
          $field = field_get_items('node', $tnode, $field_name);
          $field = $field[$row['field'][1]];

          $source_field = field_get_items('node', $node, $field_name);
          $source_field = $source_field[$row['field'][1]];

          $translated_content = trim($field['value']);
          $source_content = trim($source_field['value']);

          $field_name = implode('::', $row['field']);
          break;
      }

      $tfields[$field_name] = $this->addFieldData($tjiid, $source_content, $translated_content);
    }

    return $tfields;
  }

  /**
   * Get structured source and translated data using Entity Translation method.
   *
   * @param object $entity
   *   Source Entity.
   * @param string $tji_entity_type
   *   The Job Item source entity type.
   * @param array $data_items
   *   Job item data items.
   * @param int $tjiid
   *   Translation Job Item ID.
   * @param array $languages
   *   Source and target languages for the Job.
   *
   * @return array
   *   Translated fields data.
   */
  public function getUpdatedDataEntityTranslation($entity, $tji_entity_type, $data_items, $tjiid, $languages) {
    // Entity Translation module is enabled.
    $entity_bundle = isset($entity->type) ? $entity->type : NULL;
    $fields = array();
    $tfields = array();

    $fields_info = field_info_instances($tji_entity_type, $entity_bundle);

    if (!$entity_bundle) {
      $new = array();
      foreach ($fields_info as $value) {
        $new += $value;
      }
      $fields_info = $new;
    }

    foreach ($data_items[$tjiid] as $row) {
      $fields[] = $row['field'][0];
    }

    foreach ($fields_info as $field_info) {
      // Check if field is translatable.
      if (!in_array($field_info['field_name'], $fields)) {
        continue;
      }

      $field_name = $field_info['field_name'];
      $source_data = field_get_items($field_info['entity_type'], $entity, $field_name, $languages['source']);
      $translation_data = field_get_items($field_info['entity_type'], $entity, $field_name, $languages['target']);

      foreach ($source_data as $key => $field) {
        $field_name = $field_info['field_name'] . '::' . $key . '::value';
        $tfields[$field_name] = $this->addFieldData($tjiid, $source_data[$key]['safe_value'], $translation_data[$key]['safe_value']);
      }

    }
    return $tfields;
  }

  /**
   * Structured field data for TM Update file.
   *
   * @param int $tjiid
   *   The Job Item ID.
   * @param array $source_data
   *   Field data in Source language.
   * @param array $translation_data
   *   Field data in Target language.
   *
   * @return array
   *   Array of formatted source and translation content for each field.
   */
  public function addFieldData($tjiid, $source_data, $translation_data) {
    $field = [
      'tjiid' => $tjiid,
      'source_content' => str_replace(['“', '”'], '\"', htmlspecialchars($source_data)) . PHP_EOL,
      'target_content' => str_replace(['“', '”'], '\"', htmlspecialchars($translation_data)) . PHP_EOL,
    ];

    return $field;
  }

  /**
   * Gets status of job items from Lionbridge and downloads translated items.
   *
   * @param TMGMTJob $job
   *   The translation job object.
   *
   * @return bool
   *   The indicator of a complete translation job.
   *
   * @throws TMGMTException
   */
  public function fetchJob(TMGMTJob $job) {
    $translator = $job->getTranslator();
    $api_client = new TMGMTLionbridgeConnector($translator);
    if (!$job->reference) {
      throw new TMGMTException('Empty Job reference.');
    }

    $quote = $api_client->getQuote($job->reference);

    if (!$quote || isset($quote['Error'])) {
      throw new TMGMTException('Could not get quote.');
    }

    $translation_complete = FALSE;

    switch ($quote['Status']) {
      case LIONBRIDGE_TRANSLATION_PROVIDER_QUOTE_STATUS_ERROR:
        $job->addMessage('An error occurred with this quote.');
        $this->abortTranslation($job);
        break;

      case LIONBRIDGE_TRANSLATION_PROVIDER_QUOTE_STATUS_PENDING:
        // If this is a pending quote, add a message and a link to approve.
        $job->addMessage(
          'Quote ready for approval: <a href="@url" target="_blank">View on Lionbridge</a>', [
            '@url' => url($api_client->getEndPoint() . '/project/' . $quote['QuoteID'] . '/details'),
          ]
        );
        break;

      case LIONBRIDGE_TRANSLATION_PROVIDER_QUOTE_STATUS_CALCULATING:
        $job->addMessage('Calculating quote...');
        break;

      case LIONBRIDGE_TRANSLATION_PROVIDER_QUOTE_STATUS_AUTHORIZED:
        $job->addMessage('Translation in progress.');
        if ($translator->settings['job_type'] == 'tm_update') {
          $job->addMessage('Translation Memory Update in progress.');
        }
        break;

      case LIONBRIDGE_TRANSLATION_PROVIDER_QUOTE_STATUS_COMPLETE:
        // If it's a TM update, close job items.
        if ($translator->settings['job_type'] == 'tm_update') {
          foreach ($job->getItems() as $job_item) {
            $variables = [
              '!source' => $job_item->defaultLabel(),
              '@language' => entity_metadata_wrapper('tmgmt_job', $job)->target_language->label(),
            ];
            $job_item->accepted('Translation Memory Update of "!source" for language @language is finished.', $variables);
          }
        }
        else {
          // Get the file which contains translations for all job items.
          // Loop through the remotes, populate translated data.
          $asset_id = $quote['Projects']['Project']['Files']['File']['AssetID'];
          $data_json = $api_client->getFileTranslation($asset_id, $translator->mapToRemoteLanguage($job->target_language));
          $data = json_decode($data_json, TRUE);

          if (!$data) {
            $data_error_message = json_last_error_msg();
            throw new TMGMTException($data_error_message);
          }

          // Batch process the storing of the translated data.
          batch_set(_lionbridge_translation_provider_batch($job, $data));
          $translation_complete = TRUE;
        }
        break;
    }
    return $translation_complete;
  }

  /**
   * {@inheritdoc}
   */
  public function abortTranslation(TMGMTJob $job) {
    $translator = $job->getTranslator();
    $api_client = new TMGMTLionbridgeConnector($translator);
    $quote = $api_client->getQuote($job->reference);

    // If the status is Pending reject the quote on Lionbridge.
    if ($quote && empty($quote['Errors']) && $quote['Status'] == LIONBRIDGE_TRANSLATION_PROVIDER_QUOTE_STATUS_PENDING) {
      $api_client->rejectQuote($quote['QuoteID']);
    }

    $job->aborted();
  }

  /**
   * Sends a file to the Lionbridge translation service.
   */
  protected function addFile($translation, TMGMTLionbridgeConnector $api_client) {
    return $api_client->addFile(
      $translation['source_language'],
      $translation['fileName'],
      $translation['content_type'],
      $translation['file_content']
    );
  }

  /**
   * Sends a create project request to the Lionbridge translation service.
   */
  protected function addProject($project_data, TMGMTLionbridgeConnector $api_client) {
    return $api_client->addProject(
      $project_data['project_title'],
      $project_data['service_id'],
      $project_data['source_language'],
      $project_data['target_languages'],
      $project_data['product_assets'],
      $project_data['file_assets']
    );
  }

  /**
   * Gets a quote from the Lionbridge translation service.
   */
  protected function generateQuote($project_id, $notification_url, TMGMTLionbridgeConnector $api_client) {
    return $api_client->generateQuote($project_id, $notification_url);
  }

  /**
   * Create JSON encoded file.
   *
   * @param string $file_name
   *   The lionbridge connector.
   * @param string $source_language
   *   The source language.
   * @param mixed $data
   *   The data to be flattened.
   * @param TMGMTLionbridgeConnector $api_client
   *   The API client.
   *
   * @return array
   *   The created file.
   */
  protected function generateFile($file_name, $source_language, $data, TMGMTLionbridgeConnector $api_client) {
    $translatable_content = [];

    $data_flat = array_filter(tmgmt_flatten_data($data), '_tmgmt_filter_data');

    foreach ($data_flat as $key => $value) {
      list($tjiid, $data_item_key) = explode('][', $key, 2);

      // This is the actual text sent to Lionbridge for translation.
      $translatable_content[$tjiid][$data_item_key] = str_replace([
        '“',
        '”',
      ], '\"', $value['#text']);
    }

    $file_content = json_encode($translatable_content, JSON_PRETTY_PRINT | JSON_UNESCAPED_UNICODE);

    $translation = [
      'source_language' => $source_language,
      'fileName' => urlencode($file_name . '.json'),
      'content_type' => 'application/json',
      'file_content' => $file_content,
    ];

    return $this->addFile($translation, $api_client);
  }

}
